@@ -1,6 +1,6 @@
 ###########################################################
 # A Perl package for showing/modifying JPEG (meta)data.   #
-# Copyright (C) 2004 Stefano Bettelli                     #
+# Copyright (C) 2004,2005 Stefano Bettelli                #
 # See the COPYING and LICENSE files for license terms.    #
 ###########################################################
 package Image::MetaData::JPEG::Tables;
@@ -30,10 +30,11 @@ our %EXPORT_TAGS =                                                           #
      TagsAPP1    => [qw($APP1_EXIF_TAG $APP1_XMP_TAG $APP1_TIFF_SIG),        #
 		     qw($APP1_TH_JPEG $APP1_TH_TIFF $APP1_TH_TYPE),          #
 		     qw($THJPEG_OFFSET $THJPEG_LENGTH),                      #
-		     qw($THTIFF_OFFSET $THTIFF_LENGTH %IFD_SUBDIRS) ],       #
+		     qw($THTIFF_OFFSET $THTIFF_LENGTH),                      #
+		     qw(%IFD_SUBDIRS $HASH_MAKERNOTES $MAKERNOTE_TAG)],      #
      TagsAPP2    => [qw($APP2_FPXR_TAG $APP2_ICC_TAG)],                      #
      TagsAPP3    => [qw($APP3_EXIF_TAG %IFD_SUBDIRS)],                       #
-     TagsAPP13   => [qw($APP13_PHOTOSHOP_IPTC $APP13_PHOTOSHOP_IDENTIFIER),  #
+     TagsAPP13   => [qw($APP13_PHOTOSHOP_IPTC $APP13_PHOTOSHOP_IDS),         #
 		     qw($APP13_PHOTOSHOP_TYPE $APP13_IPTC_TAGMARKER),        #
 		     qw($APP13_PHOTOSHOP_DIRNAME $APP13_IPTC_DIRNAME)],      #
      TagsAPP14   => [qw($APP14_PHOTOSHOP_IDENTIFIER)],                       #
@@ -142,7 +143,7 @@ our $APP1_TH_JPEG		= 6;                                         #
 our $APP2_FPXR_TAG		= "FPXR\000";                                #
 our $APP2_ICC_TAG		= "ICC_PROFILE\000";                         #
 our $APP3_EXIF_TAG		= "Meta\000\000";                            #
-our $APP13_PHOTOSHOP_IDENTIFIER	= "Photoshop 3.0\000";                       #
+our $APP13_PHOTOSHOP_IDS        = ["Photoshop 3.0\000",'Adobe_Photoshop2.5:'];
 our $APP13_PHOTOSHOP_TYPE	= '8BIM';                                    #
 our $APP13_PHOTOSHOP_IPTC	= 0x0404;                                    #
 our $APP13_PHOTOSHOP_DIRNAME    = 'Photoshop_RECORDS';                       #
@@ -222,15 +223,15 @@ my $IFD_datetime = '('.$IFD_dt_full.'|    :  :     :  :  |\s{19})\000';      #
 # value can be only [2,1] (for YCbCr 4:2:2) or [2,2] (for YCbCr 4:2:0).      #
 my $SSR_YCCsampl = sub { die unless $_[0] == 2 && $_[1] =~ /1|2/; };         #
 #--- Mandatory records for IFD0 and IFD1 (not calculated) -------------------#
-my $HASH_APP1_IFD01_MANDATORY = {'XResolution'         => [1, 72],           #
-				 'YResolution'         => [1, 72],           #
-				 'ResolutionUnit'      =>  2, };             #
+my $HASH_APP1_IFD01_MANDATORY = {'XResolution'               => [1, 72],     #
+				 'YResolution'               => [1, 72],     #
+				 'ResolutionUnit'            =>  2, };       #
 my $HASH_APP1_IFD0_MANDATORY  = {%$HASH_APP1_IFD01_MANDATORY,                #
-				 'YCbCrPositioning'    =>  1, };             #
+				 'YCbCrPositioning'          =>  1, };       #
 my $HASH_APP1_IFD1_MANDATORY  = {%$HASH_APP1_IFD01_MANDATORY,                #
-				 'YCbCrSubSampling'    => [2,1],             #
-				 'PhotometricInterpretation' => 2,           #
-				 'PlanarConfiguration' =>  1, };             #
+				 'YCbCrSubSampling'          => [2, 1],      #
+				 'PhotometricInterpretation' =>  2,          #
+				 'PlanarConfiguration'       =>  1, };       #
 #--- Legal records' list ----------------------------------------------------#
 my $HASH_APP1_IFD01_MAIN =                                                   #
 {0x0100 => ['ImageWidth',                 $LONG,      1, $IFD_integer, 'T'], #
@@ -324,6 +325,12 @@ my $HASH_APP1_IFD01_ADDITIONAL =                                             #
  0x0209 => ['JPEGACTables',               $LONG,  undef, 'invalid'    ],     #
  0x02bc => ['XML_Packet',                 $BYTE,  undef, $IFD_integer ], };  #
 #----------------------------------------------------------------------------#
+# The following company-related fields are marked as invalid because they    #
+# are present also in the SubIFD section (with different numerical values)   #
+# and I don't want the two entries to collide when setting IMAGE_DATA:       #
+# 'FlashEnergy', 'SpatialFrequencyResponse', FocalPlane[XY]Resolution',      #
+# 'FocalPlaneResolutionUnit', 'ExposureIndex', 'SensingMethod', 'CFAPattern' #
+#----------------------------------------------------------------------------#
 my $HASH_APP1_IFD01_COMPANIES =                                              #
 {0x800d => ['ImageID',                    $ASCII, undef, $IFD_Cstring ],     #
  0x80b9 => ['RefPts',                     undef,  undef, 'invalid'    ],     #
@@ -343,7 +350,7 @@ my $HASH_APP1_IFD01_COMPANIES =                                              #
  0x821a => ['MatrixWorldToCamera',        $FLOAT,    16, $IFD_float   ],     #
  0x827d => ['WriterSerialNumber',         undef,  undef, 'invalid'    ],     #
  0x828d => ['CFARepeatPatternDim',        $SHORT,     2, $IFD_integer ],     #
- 0x828e => ['CFAPattern',                 $BYTE,  undef, $IFD_integer ],     #
+ 0x828e => ['CFAPattern',                 $BYTE,  undef, 'invalid'    ],     #
  0x828f => ['BatteryLevel',               $ASCII, undef, $IFD_Cstring ],     #
  0x830e => ['ModelPixelScaleTag',         $DOUBLE,    3, $IFD_float   ],     #
  0x83bb => ['IPTC/NAA',                   $ASCII, undef, $IFD_Cstring ],     #
@@ -384,18 +391,18 @@ my $HASH_APP1_IFD01_COMPANIES =                                              #
  0x885d => ['FaxSubAddress',              $ASCII, undef, $IFD_Cstring ],     #
  0x885e => ['FaxRecvTime',                $LONG,      1, $IFD_integer ],     #
  0x8871 => ['FedExEDR',                   undef,  undef, 'invalid'    ],     #
- 0x920b => ['FlashEnergy',                $RATIONAL,undef,$IFD_integer],     #
+ 0x920b => ['FlashEnergy',               $RATIONAL,undef,'invalid'    ],     #
  0x920c => ['SpatialFrequencyResponse',   undef,  undef, 'invalid'    ],     #
  0x920d => ['Noise',                      undef,  undef, 'invalid'    ],     #
- 0x920e => ['FocalPlaneXResolution',      $RATIONAL,  1, $IFD_integer ],     #
- 0x920f => ['FocalPlaneYResolution',      $RATIONAL,  1, $IFD_integer ],     #
- 0x9210 => ['FocalPlaneResolutionUnit',   $SHORT,     1, $IFD_integer ],     #
+ 0x920e => ['FocalPlaneXResolution',      $RATIONAL,  1, 'invalid'    ],     #
+ 0x920f => ['FocalPlaneYResolution',      $RATIONAL,  1, 'invalid'    ],     #
+ 0x9210 => ['FocalPlaneResolutionUnit',   $SHORT,     1, 'invalid'    ],     #
  0x9211 => ['ImageNumber',                $LONG,      1, $IFD_integer ],     #
  0x9212 => ['SecurityClassification',     $ASCII, undef, $IFD_Cstring ],     #
  0x9213 => ['ImageHistory',               $ASCII, undef, $IFD_Cstring ],     #
- 0x9215 => ['ExposureIndex',              $RATIONAL,undef,$IFD_integer],     #
+ 0x9215 => ['ExposureIndex',             $RATIONAL,undef,'invalid'    ],     #
  0x9216 => ['TIFF/EPStandardID',          $BYTE,      4, $IFD_integer ],     #
- 0x9217 => ['SensingMethod',              $SHORT,     1, $IFD_integer ],     #
+ 0x9217 => ['SensingMethod',              $SHORT,     1, 'invalid'    ],     #
  0x923f => ['StoNits',                    $DOUBLE,    1, $IFD_float   ],     #
  0x935c => ['ImageSourceData',            undef,  undef, 'invalid'    ],     #
  0xc4a5 => ['PrintIM_Data',               undef,  undef, 'invalid'    ],     #
@@ -427,6 +434,7 @@ my $IFD_flash    = '([01579]|1[356]|2[459]|3[12]|6[59]|7[1379]|89|9[35])';   #
 my $IFD_hexstr   = '[0-9a-fA-F]+\000+';             # hexadecimal ASCII str  #
 my $IFD_Exifver  = '0(100|110|200|210|220|221)';    # known Exif versions    #
 my $IFD_setdesc  = '.{4}(\376\377(.{2})*\000\000)*'; # for DeviceSettingDesc.#
+my $IFD_compconf = '(\004\005\006|\001\002\003)\000';# for ComponentsConfig. #
 #--- Special screen rules ---------------------------------------------------#
 # a SubjectArea tag indicates the location and area of the main subject. The #
 # tag can contain 2, 3 or 4 integer numbers (see Exif 2.2 for their meaning) #
@@ -439,13 +447,12 @@ my $SSR_cfapattern  = sub { my ($x, $y) = unpack 'nn', $_[0];                #
 			    die if length $_[0] != 4+$x*$y;                  #
 			    die if $_[0] !~ /^.{4}[0-6]*$/; };               #
 #--- Mandatory records ------------------------------------------------------#
-my $gdim = sub {$_[1]->{parent}?($_[1]->{parent}->get_dimensions())[$_[0]]:0};
-my $HASH_APP1_SUBIFD_MANDATORY = {'ExifVersion'             => '0220',       #
-				  'ComponentsConfiguration' => '1230',       #
-				  'FlashpixVersion'         => '0100',       #
-				  'ColorSpace'              => 1,            #
-				  'PixelXDimension' => sub {&$gdim(0,@_)},   #
-				  'PixelYDimension' => sub {&$gdim(1,@_)} }; #
+my $HASH_APP1_SUBIFD_MANDATORY = {'ExifVersion'       => '0220',             #
+ 			    'ComponentsConfiguration' => "\001\002\003\000", #
+				  'FlashpixVersion'   => '0100',             #
+				  'ColorSpace'        => 1,                  #
+				  'PixelXDimension'   => 0,   # global info  #
+				  'PixelYDimension'   => 0 }; # needed here! #
 #--- Legal records' list ----------------------------------------------------#
 my $HASH_APP1_SUBIFD_GENERAL =                                               #
 {0x829a => ['ExposureTime',               $RATIONAL,  1, $IFD_integer    ],  #
@@ -457,7 +464,7 @@ my $HASH_APP1_SUBIFD_GENERAL =                                               #
  0x9000 => ['ExifVersion',                $UNDEF,     4, $IFD_Exifver    ],  #
  0x9003 => ['DateTimeOriginal',           $ASCII,    20, $IFD_datetime   ],  #
  0x9004 => ['DateTimeDigitized',          $ASCII,    20, $IFD_datetime   ],  #
- 0x9101 => ['ComponentsConfiguration',    $UNDEF,     4, '(456|123)0'    ],  #
+ 0x9101 => ['ComponentsConfiguration',    $UNDEF,     4, $IFD_compconf   ],  #
  0x9102 => ['CompressedBitsPerPixel',     $RATIONAL,  1, $IFD_integer    ],  #
  0x9201 => ['ShutterSpeedValue',          $SRATIONAL, 1, $IFD_signed     ],  #
  0x9202 => ['ApertureValue',              $RATIONAL,  1, $IFD_integer    ],  #
@@ -470,7 +477,7 @@ my $HASH_APP1_SUBIFD_GENERAL =                                               #
  0x9209 => ['Flash',                      $SHORT,     1, $IFD_flash      ],  #
  0x920a => ['FocalLength',                $RATIONAL,  1, $IFD_integer    ],  #
  0x9214 => ['SubjectArea',                $SHORT, undef, $SSR_subjectarea],  #
- 0x927c => ['MakerNote',                  $UNDEF, undef, '.*'            ],  #
+ 0x927c => ['MakerNote',                  $UNDEF, undef, 'invalid'       ],  #
  0x9286 => ['UserComment',                $UNDEF, undef, $IFD_Ustring    ],  #
  0x9290 => ['SubSecTime',                 $ASCII, undef, $IFD_subsecs    ],  #
  0x9291 => ['SubSecTimeOriginal',         $ASCII, undef, $IFD_subsecs    ],  #
@@ -489,8 +496,8 @@ my $HASH_APP1_SUBIFD_GENERAL =                                               #
  0xa214 => ['SubjectLocation',            $SHORT,     2, $IFD_integer    ],  #
  0xa215 => ['ExposureIndex',              $RATIONAL,  1, $IFD_integer    ],  #
  0xa217 => ['SensingMethod',              $SHORT,     1, '[1-578]'       ],  #
- 0xa300 => ['FileSource',                 $UNDEF,     1, '3'             ],  #
- 0xa301 => ['SceneType',                  $UNDEF,     1, '1'             ],  #
+ 0xa300 => ['FileSource',                 $UNDEF,     1, '\003'          ],  #
+ 0xa301 => ['SceneType',                  $UNDEF,     1, '\001'          ],  #
  0xa302 => ['CFAPattern',                 $UNDEF, undef, $SSR_cfapattern ],  #
  0xa401 => ['CustomRendered',             $SHORT,     1, '[01]'          ],  #
  0xa402 => ['ExposureMode',               $SHORT,     1, '[012]'         ],  #
@@ -519,7 +526,7 @@ my $HASH_APP1_SUBIFD_GENERAL =                                               #
  0xfe55 => ['_Saturation',    $ASCII, undef, 'Saturation: .*\000'        ],  #
  0xfe56 => ['_Sharpness',     $ASCII, undef, 'Sharpness: .*\000'         ],  #
  0xfe57 => ['_Smoothness',    $ASCII, undef, 'Smoothness: .*\000'        ],  #
- 0xfe58 => ['_MoireFilter',   $ASCII, undef, 'Moire Filter: .*\000'     ],}; #
+ 0xfe58 => ['_MoireFilter',   $ASCII, undef, 'Moire Filter: .*\000'   ], };  #
 #============================================================================#
 #============================================================================#
 #============================================================================#
@@ -534,11 +541,11 @@ my $HASH_INTEROP_MANDATORY = {'InteroperabilityVersion' => '0100',           #
 			      'InteroperabilityIndex'   => 'R98'  };         #
 #--- Legal records' list ----------------------------------------------------#
 my $HASH_INTEROP_GENERAL =                                                   #
-{0x0001 => ['InteroperabilityIndex',      $ASCII,     4, 'R98\000'    ],     #
- 0x0002 => ['InteroperabilityVersion',    $UNDEF,     4, '[0-9]{4}'   ],     #
- 0x1000 => ['RelatedImageFileFormat',     $ASCII, undef, '.*\000'     ],     #
- 0x1001 => ['RelatedImageWidth',          $LONG,      1, '[0-9]*'     ],     #
- 0x1002 => ['RelatedImageLength',         $LONG,      1, '[0-9]*'     ], };  #
+{0x0001 => ['InteroperabilityIndex',      $ASCII,     4, 'R98\000'     ],    #
+ 0x0002 => ['InteroperabilityVersion',    $UNDEF,     4, '[0-9]{4}'    ],    #
+ 0x1000 => ['RelatedImageFileFormat',     $ASCII, undef, $IFD_Cstring  ],    #
+ 0x1001 => ['RelatedImageWidth',          $LONG,      1, '[0-9]*'      ],    #
+ 0x1002 => ['RelatedImageLength',         $LONG,      1, '[0-9]*'      ], }; #
 #============================================================================#
 #============================================================================#
 #============================================================================#
@@ -726,7 +733,7 @@ my $HASH_IPTC_GENERAL =                                                      #
  20  => ['SupplementalCategory',        ' ', 1, 32, $IPTC_re_line         ], #
  22  => ['FixtureIdentifier',           'N', 1, 32, $IPTC_re_word         ], #
  25  => ['Keywords',                    ' ', 1, 64, $IPTC_re_line         ], #
- 26  => ['ContentLocationCode',         ' ', 3,  3, '[A-Z]{3}?'           ], #
+ 26  => ['ContentLocationCode',         ' ', 3,  3, '[A-Z]{3}'            ], #
  27  => ['ContentLocationName',         ' ', 1, 64, $IPTC_re_line         ], #
  30  => ['ReleaseDate',                 'N', 8,  8, $IPTC_re_date         ], #
  35  => ['ReleaseTime',                 'N',11, 11, $IPTC_re_time         ], #
@@ -871,6 +878,17 @@ $$HASH_APP1_SUBIFD{Interop} = $HASH_APP1_INTEROP;  # Interoperability tags   #
 #============================================================================#
 #============================================================================#
 #============================================================================#
+# MakerNote stuff is stored in a separated file; the return value of this    #
+# inclusion is the $HASH_MAKERNOTES hash reference, containing all relevant  #
+# parameters. We only have to link this new table into $HASH_APP1_SUBIFD.    #
+#----------------------------------------------------------------------------#
+our $HASH_MAKERNOTES = require 'Image/MetaData/JPEG/Tables_makernotes.pl';   #
+$$HASH_APP1_SUBIFD{'MakerNoteData_' . $_} =                                  #
+    generate_lookup($$HASH_MAKERNOTES{$_}{tags} ,0)                          #
+    for keys %$HASH_MAKERNOTES;                                              #
+#============================================================================#
+#============================================================================#
+#============================================================================#
 # The following hash is the database for the tag-to-tagname translation; of  #
 # course, records with a textual tag are not listed here. The navigation     #
 # through this structure is best done with the help of the JPEG_lookup       #
@@ -914,12 +932,16 @@ sub JPEG_lookup {
     # print a debugging message and return immediately unless
     # all arguments are scalars (i.e., references are not allowed)
     for (@_) { print "wrong argument(s) in JPEG_lookup call", return if ref; }
-    # join all arguments with '@'
+    # delete all undefined or "false" arguments
+    @_ = grep { defined $_ } @_;
+    # join all remaining arguments
     my $keystring = join('@', @_);
     # split the resulting string on '@'
     my @keylist = split('@', $keystring);
     # extract and save the last argument for special treatment
     my $last = pop @keylist;
+    # delete all false arguments
+    @keylist = grep { $_ } @keylist;
     # refuse to work with $last undefined
     return unless defined $last;
     # consume the list of "normal" arguments: they must be successive
@@ -978,6 +1000,7 @@ our %IFD_SUBDIRS =                                                           #
 			'__mandatory'        => $HASH_GPS_MANDATORY },       #
  'APP1@IFD0@SubIFD' => {'__syntax'           => $HASH_APP1_SUBIFD_GENERAL,   #
 			'__mandatory'        => $HASH_APP1_SUBIFD_MANDATORY, #
+			'MakerNote'          => 'MakerNoteData',             #
 			'InteroperabilityOffset' => 'Interop'},              #
  'APP1@IFD0@SubIFD@Interop' => {'__syntax'   => $HASH_INTEROP_GENERAL,       #
 				'__mandatory'=> $HASH_INTEROP_MANDATORY },   #
@@ -1002,9 +1025,8 @@ our $THJPEG_OFFSET = JPEG_lookup('APP1@IFD1@JPEGInterchangeFormat');         #
 our $THJPEG_LENGTH = JPEG_lookup('APP1@IFD1@JPEGInterchangeFormatLength');   #
 our $THTIFF_OFFSET = JPEG_lookup('APP1@IFD1@StripOffsets');                  #
 our $THTIFF_LENGTH = JPEG_lookup('APP1@IFD1@StripByteCounts');               #
+our $MAKERNOTE_TAG = JPEG_lookup('APP1@IFD0@SubIFD@MakerNote');              #
 #----------------------------------------------------------------------------#
 
-
-
 # successful package load
 1;
